"""Utilities for interacting with database through the Inventory API

Until assets are created entirely in the database, this script
provides a bridge between the file-based project inventory and configuration.

Usage:
    $ # Create new project:
    $ python -m avalon.inventory --init
    $ python -m avalon.inventory --save

    $ # Manage existing project
    $ python -m avalon.inventory --load
    $ # Update the .inventory.toml or .config.toml
    $ python -m avalon.inventory --save

"""

import os
import sys
import copy

from avalon import schema, io
from avalon.vendor import toml

self = sys.modules[__name__]
self.collection = None

__all__ = [
    "init",
    "save",
    "load",
]


DEFAULTS = {
    "config": {
        "schema": "avalon-core:config-1.0",
        "apps": [
            {
                "name": "shell",
                "label": "Shell"
            },
            {
                "name": "maya2016",
                "label": "Autodesk Maya 2016"
            },
            {
                "name": "nuke10",
                "label": "The Foundry Nuke 10.0"
            },
        ],
        "tasks": [
            {"name": "model"},
            {"name": "render"},
            {"name": "animate"},
            {"name": "rig"},
            {"name": "lookdev"},
            {"name": "layout"},
        ],
        "template": {
            "work":
                "{root}/{project}/{silo}/{asset}/work/"
                "{task}/{app}",
            "publish":
                "{root}/{project}/{silo}/{asset}/publish/"
                "{subset}/v{version:0>3}/{subset}.{representation}"
        }
    },
    "inventory": {
        "schema": "avalon-core:inventory-1.0",
        "assets": [
            {
                "name": "hero",
                "label": "Hero"
            },
            {
                "name": "villain"
            }
        ],
        "film": [
            {
                "name": "shot1",
                "frameStart": 1000,
                "frameEnd": 1143
            },
            {
                "name": "shot2",
                "frameStart": 1000,
                "frameEnd": 1081
            },
        ]
    }
}


def create_project(name):
    if io.find_one({"type": "project", "name": name}):
        raise RuntimeError("%s already exists" % name)

    return io.insert_one({
        "schema": "avalon-core:project-2.0",
        "type": "project",
        "name": name,
        "data": dict(),
        "config": DEFAULTS["config"],
        "parent": None,
    }).inserted_id


def create_asset(name, silo, data, parent):
    assert isinstance(parent, io.ObjectId)
    if io.find_one({"type": "asset", "name": name}):
        raise RuntimeError("%s already exists" % name)

    return io.insert_one({
        "schema": "avalon-core:asset-2.0",
        "name": name,
        "silo": silo,
        "parent": parent,
        "type": "asset",
        "data": data
    }).inserted_id


def save(name, config, inventory):
    """Write `config` and `inventory` to database as `name`

    Given a configuration and inventory, this function writes
    the changes to the current database.

    Arguments:
        name (str): Project name
        config (dict): Current configuration
        inventory (dict): Current inventory

    """

    config = copy.deepcopy(config)
    inventory = copy.deepcopy(inventory)

    if "config" not in config.get("schema", ""):
        raise schema.SchemaError("Missing schema for config")

    if "inventory" not in inventory.get("schema", ""):
        raise schema.SchemaError("Missing schema for inventory")

    handlers = {
        "avalon-core:inventory-1.0": _save_inventory_1_0,
        "avalon-core:config-1.0": _save_config_1_0
    }

    for data in (inventory, config):
        try:
            schema_ = data.get("schema")
            handler = handlers[schema_]

        except KeyError:
            raise schema.SchemaError(
                "ERROR: Missing handler for %s)" % (schema))

        else:
            schema.validate(data)
            print("Saving %s.." % schema_)
            handler(name, data)


def init(name):
    """Initialise a new project called `name`

    Upon creating a new project, call init to establish an example
    inventory and configuration for your project.

    Arguments:
        name (str): Name of new project

    """

    project = io.find_one({"type": "project"})

    if project is not None:
        raise Exception("'%s' already exists, "
                        "try --load instead" % project["name"])

    config = DEFAULTS["config"]
    inventory = DEFAULTS["inventory"]

    return config, inventory


def load(name):
    """Read project called `name` from database

    Arguments:
        name (str): Project name

    """

    print("Loading .inventory.toml and .config.toml..")

    project = io.find_one({"type": "project"})

    if project is None:
        msg = "'{0}' not found, try --init to start a new project".format(name)

        projects = ""
        for project in io.projects():
            projects += "\n- {0}".format(project["name"])

        if projects:
            msg += (
                ", or load a project from the database.\nProjects:{1}".format(
                    name, projects
                )
            )
        else:
            msg += "."

        raise Exception(msg)

    else:
        config = project["config"]
        inventory = {"schema": "avalon-core:inventory-1.0"}
        for asset in io.find({"type": "asset", "parent": project["_id"]}):
            silo = asset["silo"]
            data = asset["data"]

            data.pop("visualParent", None)  # Hide from manual editing

            if silo not in inventory:
                inventory[silo] = list()

            inventory[silo].append(dict(data, **{"name": asset["name"]}))

            for key, value in project["data"].items():
                inventory[key] = value

    config = dict(
        DEFAULTS["config"],
        **config
    )

    return config, inventory


def ls():
    """Return a list of project names in database."""

    return [project["name"] for project in io.projects()]


def _save_inventory_1_0(project_name, data):
    data.pop("schema")

    # Separate project metadata from assets
    metadata = {}
    for key, value in data.copy().items():
        if not isinstance(value, list):
            print("Separating project metadata: %s" % key)
            metadata[key] = data.pop(key)

    _filter = {"type": "project"}

    document = io.find_one(_filter)
    if document is None:
        print("'%s' not found, creating.." % project_name)
        _id = create_project(project_name)

        document = io.find_one({"_id": _id})

    print("Updating project data..")
    for key, value in metadata.items():
        document["data"][key] = value

    io.replace_one(_filter, document)

    print("Updating assets..")
    added = list()
    updated = list()
    missing = list()
    for silo, assets in data.items():
        for asset in assets:

            _filter = {
                "name": asset["name"],
                "type": "asset",
            }

            asset_doc = io.find_one(_filter)

            if asset_doc is None:
                asset["silo"] = silo
                missing.append(asset)
                continue

            for key, value in asset.items():
                asset_doc["data"][key] = value

                if key not in asset_doc["data"]:
                    added.append("%s.%s: %s" % (asset["name"], key, value))

                elif asset_doc["data"][key] != value:
                    updated.append("%s.%s: %s -> %s" % (asset["name"],
                                                        key,
                                                        asset_doc["data"][key],
                                                        value))

            io.replace_one(_filter, asset_doc)

    for data in missing:
        print("+ added %s" % data["name"])

        create_asset(
            name=data.pop("name"),
            silo=data.pop("silo"),
            data=data,
            parent=document["_id"]
        )

    else:
        print("| nothing missing")

    _report(added, missing)


def _save_config_1_0(project_name, data):
    _filter = {"type": "project"}

    document = io.find_one(_filter)

    config = document["config"]

    config["apps"] = data.get("apps", [])
    config["tasks"] = data.get("tasks", [])
    config["template"].update(data.get("template", {}))
    config["families"] = data.get("families", [])
    config["groups"] = data.get("groups", [])

    schema.validate(document)

    io.replace_one(_filter, document)


def _report(added, updated):
    if added:
        print("+ added:")
        print("\n".join(" - %s" % item for item in added))
    else:
        print("| nothing added")

    if updated:
        print("+ updated:")
        print("\n".join("  %s" % item for item in updated))
    else:
        print("| already up to date")


def _read(root, name):
    fname = os.path.join(root, ".%s.toml" % name)

    try:
        with open(fname) as f:
            data = toml.load(f)
    except IOError:
        raise

    return data


def _write(root, name, data):
    fname = os.path.join(root, ".%s.toml" % name)

    try:
        with open(fname, "w") as f:
            toml.dump(data, f)
            schema.validate(data)
    except IOError:
        raise

    return data


def _cli():
    import argparse

    parser = argparse.ArgumentParser(__doc__)

    parser.add_argument("--silo",
                        help="Optional container of silos",
                        action="append",
                        default=["assets", "film"])
    parser.add_argument("--silo-parent", help="Optional silo silo_parent")
    parser.add_argument("--init",
                        action="store_true",
                        help="Start a new project")
    parser.add_argument("--save",
                        action="store_true",
                        help="Save inventory from disk to database")
    parser.add_argument("--load",
                        nargs="?",
                        default=False,
                        help="Load inventory from database to disk")
    parser.add_argument("--ls",
                        action="store_true",
                        help="List all projects in database")
    parser.add_argument("--extract",
                        action="store_true",
                        help="Generate config and inventory "
                             "from assets already on disk.")
    parser.add_argument("--overwrite",
                        action="store_true",
                        help="Overwrite exitsing assets on upload")
    parser.add_argument("--upload",
                        action="store_true",
                        help="Upload generated project to database.")
    parser.add_argument("--root", help="Absolute path to project.")

    kwargs = parser.parse_args()

    root = kwargs.root or os.getcwd()
    name = kwargs.load or os.path.basename(root)

    if any([kwargs.load,
            kwargs.save,
            kwargs.upload,
            kwargs.init,
            kwargs.ls]) or kwargs.load is None:
        os.environ["AVALON_PROJECT"] = name
        io.install()

    if kwargs.init:
        config, inventory = init(name)
        _write(root, "config", config)
        _write(root, "inventory", inventory)
        print("Success!")

    elif kwargs.load or kwargs.load is None:
        config, inventory = load(name)
        _write(root, "config", config)
        _write(root, "inventory", inventory)
        print("Success!")

    elif kwargs.save:
        inventory = _read(root, "inventory")
        config = _read(root, "config")
        save(name, config, inventory)
        print("Successfully saved to %s" % os.getenv("AVALON_MONGO"))

    elif kwargs.ls:
        msg = "Projects in database:"
        for name in ls():
            msg += "\n- {0}".format(name)
        print(msg)

    else:
        print(__doc__)


if __name__ == '__main__':
    try:
        _cli()
    except Exception as e:
        print(e)
        sys.exit(1)



"""Public API

Anything that isn't defined here is INTERNAL and unreliable for external use.

"""

from .pipeline import (
    install,

    ls,

    imprint_container,
    parse_container,

    get_current_comp,
    comp_lock_and_undo_chunk

)

from .workio import (
    open_file,
    save_file,
    current_file,
    has_unsaved_changes,
    file_extensions,
    work_root
)

from .lib import (
    maintained_selection
)

__all__ = [
    "install",

    "ls",

    "imprint_container",
    "parse_container",

    "get_current_comp",
    "comp_lock_and_undo_chunk",

    # Workfiles API
    "open_file",
    "save_file",
    "current_file",
    "has_unsaved_changes",
    "file_extensions",
    "work_root",

    "maintained_selection"

]

# Backwards API compatibility
open = open_file
save = save_file



import os
import logging

from ... import style
from ...vendor.Qt import QtCore
from ...vendor import qtawesome

from ..models import TreeModel, Item

log = logging.getLogger(__name__)


class FilesModel(TreeModel):
    """Model listing files with specified extensions in a root folder"""
    Columns = ["filename", "date"]

    FileNameRole = QtCore.Qt.UserRole + 2
    DateModifiedRole = QtCore.Qt.UserRole + 3
    FilePathRole = QtCore.Qt.UserRole + 4
    IsEnabled = QtCore.Qt.UserRole + 5

    def __init__(self, file_extensions, parent=None):
        super(FilesModel, self).__init__(parent=parent)

        self._root = None
        self._file_extensions = file_extensions
        self._icons = {"file": qtawesome.icon("fa.file-o",
                                              color=style.colors.default)}

    def set_root(self, root):
        self._root = root
        self.refresh()

    def _add_empty(self):

        item = Item()
        item.update({
            # Put a display message in 'filename'
            "filename": "No files found.",
            # Not-selectable
            "enabled": False,
            "filepath": None
        })

        self.add_child(item)

    def refresh(self):

        self.clear()
        self.beginResetModel()

        root = self._root

        if not root:
            self.endResetModel()
            return

        if not os.path.exists(root):
            # Add Work Area does not exist placeholder
            log.debug("Work Area does not exist: %s", root)
            message = "Work Area does not exist. Use Save As to create it."
            item = Item({
                "filename": message,
                "date": None,
                "filepath": None,
                "enabled": False,
                "icon": qtawesome.icon("fa.times",
                                       color=style.colors.mid)
            })
            self.add_child(item)
            self.endResetModel()
            return

        extensions = self._file_extensions

        for f in os.listdir(root):
            path = os.path.join(root, f)
            if os.path.isdir(path):
                continue

            if extensions and os.path.splitext(f)[1] not in extensions:
                continue

            modified = os.path.getmtime(path)

            item = Item({
                "filename": f,
                "date": modified,
                "filepath": path
            })

            self.add_child(item)

        self.endResetModel()

    def data(self, index, role):

        if not index.isValid():
            return

        if role == QtCore.Qt.DecorationRole:
            # Add icon to filename column
            item = index.internalPointer()
            if index.column() == 0:
                if item["filepath"]:
                    return self._icons["file"]
                else:
                    return item.get("icon", None)
        if role == self.FileNameRole:
            item = index.internalPointer()
            return item["filename"]
        if role == self.DateModifiedRole:
            item = index.internalPointer()
            return item["date"]
        if role == self.FilePathRole:
            item = index.internalPointer()
            return item["filepath"]
        if role == self.IsEnabled:
            item = index.internalPointer()
            return item.get("enabled", True)

        return super(FilesModel, self).data(index, role)

    def headerData(self, section, orientation, role):

        # Show nice labels in the header
        if role == QtCore.Qt.DisplayRole and \
                orientation == QtCore.Qt.Horizontal:
            if section == 0:
                return "Name"
            elif section == 1:
                return "Date modified"

        return super(FilesModel, self).headerData(section, orientation, role)



import sys
import time

from ...vendor.Qt import QtWidgets, QtCore
from ... import api, io, style

from ..models import AssetModel
from ..widgets import AssetWidget
from .. import lib

from .widgets import SubsetWidget, VersionWidget, FamilyListWidget

module = sys.modules[__name__]
module.window = None

# Custom roles
DocumentRole = AssetModel.DocumentRole


class Window(QtWidgets.QDialog):
    """Asset loader interface"""

    def __init__(self, parent=None):
        super(Window, self).__init__(parent)
        self.setWindowTitle(
            "Asset Loader 2.1 - %s/%s" % (
                api.registered_root(),
                api.Session.get("AVALON_PROJECT")))

        # Enable minimize and maximize for app
        self.setWindowFlags(QtCore.Qt.Window)
        self.setFocusPolicy(QtCore.Qt.StrongFocus)

        body = QtWidgets.QWidget()
        footer = QtWidgets.QWidget()
        footer.setFixedHeight(20)

        container = QtWidgets.QWidget()

        assets = AssetWidget(silo_creatable=False)
        families = FamilyListWidget()
        subsets = SubsetWidget()
        version = VersionWidget()

        # Create splitter to show / hide family filters
        asset_filter_splitter = QtWidgets.QSplitter()
        asset_filter_splitter.setOrientation(QtCore.Qt.Vertical)
        asset_filter_splitter.addWidget(assets)
        asset_filter_splitter.addWidget(families)
        asset_filter_splitter.setStretchFactor(0, 65)
        asset_filter_splitter.setStretchFactor(1, 35)

        container_layout = QtWidgets.QHBoxLayout(container)
        container_layout.setContentsMargins(0, 0, 0, 0)
        split = QtWidgets.QSplitter()
        split.addWidget(asset_filter_splitter)
        split.addWidget(subsets)
        split.addWidget(version)
        split.setSizes([180, 950, 200])

        container_layout.addWidget(split)

        body_layout = QtWidgets.QHBoxLayout(body)
        body_layout.addWidget(container)
        body_layout.setContentsMargins(0, 0, 0, 0)

        message = QtWidgets.QLabel()
        message.hide()

        footer_layout = QtWidgets.QVBoxLayout(footer)
        footer_layout.addWidget(message)
        footer_layout.setContentsMargins(0, 0, 0, 0)

        layout = QtWidgets.QVBoxLayout(self)
        layout.addWidget(body)
        layout.addWidget(footer)

        self.data = {
            "widgets": {"families": families},
            "model": {
                "assets": assets,
                "subsets": subsets,
                "version": version,
            },
            "label": {
                "message": message,
            },
            "state": {
                "template": None,
                "locations": list(),
                "context": {
                    "root": None,
                    "project": None,
                    "asset": None,
                    "assetId": None,
                    "silo": None,
                    "subset": None,
                    "version": None,
                    "representation": None,
                },
            }
        }

        families.active_changed.connect(subsets.set_family_filters)
        assets.selection_changed.connect(self.on_assetschanged)
        subsets.active_changed.connect(self.on_subsetschanged)
        subsets.version_changed.connect(self.on_versionschanged)

        # Defaults
        self.resize(1330, 700)

    # -------------------------------
    # Delay calling blocking methods
    # -------------------------------

    def refresh(self):
        self.echo("Fetching results..")
        lib.schedule(self._refresh, 50, channel="mongo")

    def on_assetschanged(self, *args):
        self.echo("Fetching asset..")
        lib.schedule(self._assetschanged, 50, channel="mongo")

    def on_subsetschanged(self, *args):
        self.echo("Fetching subset..")
        lib.schedule(self._versionschanged, 50, channel="mongo")

    def on_versionschanged(self, *args):
        self.echo("Fetching version..")
        lib.schedule(self._versionschanged, 150, channel="mongo")

    def set_context(self, context, refresh=True):
        self.echo("Setting context: {}".format(context))
        lib.schedule(lambda: self._set_context(context, refresh=refresh),
                     50, channel="mongo")

    # ------------------------------

    def _refresh(self):
        """Load assets from database"""

        # Ensure a project is loaded
        project = io.find_one({"type": "project"})
        assert project, "This is a bug"

        assets_model = self.data["model"]["assets"]
        assets_model.refresh()
        assets_model.setFocus()

        families = self.data["widgets"]["families"]
        families.refresh()

        # Update state
        state = self.data["state"]
        state["template"] = project["config"]["template"]["publish"]
        state["context"]["root"] = api.registered_root()
        state["context"]["project"] = project["name"]

    def _assetschanged(self):
        """Selected assets have changed"""

        assets_model = self.data["model"]["assets"]
        subsets = self.data["model"]["subsets"]
        subsets_model = subsets.model
        subsets_model.clear()

        t1 = time.time()

        asset_item = assets_model.get_active_index()
        if asset_item is None or not asset_item.isValid():
            return

        document = asset_item.data(DocumentRole)
        subsets_model.set_asset(document["_id"])

        # Clear the version information on asset change
        self.data["model"]["version"].set_version(None)

        self.data["state"]["context"]["asset"] = document["name"]
        self.data["state"]["context"]["assetId"] = document["_id"]
        self.data["state"]["context"]["silo"] = document["silo"]
        self.echo("Duration: %.3fs" % (time.time() - t1))

    def _versionschanged(self):

        subsets = self.data["model"]["subsets"]
        selection = subsets.view.selectionModel()

        # Active must be in the selected rows otherwise we
        # assume it's not actually an "active" current index.
        version = None
        active = selection.currentIndex()
        if active:
            rows = selection.selectedRows(column=active.column())
            if active in rows:
                node = active.data(subsets.model.ItemRole)
                if node is not None and not node.get("isGroup"):
                    version = node["version_document"]["_id"]

        self.data["model"]["version"].set_version(version)

    def _set_context(self, context, refresh=True):
        """Set the selection in the interface using a context.

        The context must contain `silo` and `asset` data by name.

        Note: Prior to setting context ensure `refresh` is triggered so that
              the "silos" are listed correctly, aside from that setting the
              context will force a refresh further down because it changes
              the active silo and asset.

        Args:
            context (dict): The context to apply.

        Returns:
            None

        """

        silo = context.get("silo", None)
        if silo is None:
            return

        asset = context.get("asset", None)
        if asset is None:
            return

        if refresh:
            # Workaround:
            # Force a direct (non-scheduled) refresh prior to setting the
            # asset widget's silo and asset selection to ensure it's correctly
            # displaying the silo tabs. Calling `window.refresh()` and directly
            # `window.set_context()` the `set_context()` seems to override the
            # scheduled refresh and the silo tabs are not shown.
            self._refresh()

        asset_widget = self.data["model"]["assets"]
        asset_widget.set_silo(silo)
        asset_widget.select_assets([asset], expand=True)

    def echo(self, message):
        widget = self.data["label"]["message"]
        widget.setText(str(message))
        widget.show()
        print(message)

        lib.schedule(widget.hide, 5000, channel="message")

    def closeEvent(self, event):
        # Kill on holding SHIFT
        modifiers = QtWidgets.QApplication.queryKeyboardModifiers()
        shift_pressed = QtCore.Qt.ShiftModifier & modifiers

        if shift_pressed:
            print("Force quitted..")
            self.setAttribute(QtCore.Qt.WA_DeleteOnClose)

        # Kill on holding SHIFT
        modifiers = QtWidgets.QApplication.queryKeyboardModifiers()
        shift_pressed = QtCore.Qt.ShiftModifier & modifiers

        if shift_pressed:
            print("Force quitted..")
            self.setAttribute(QtCore.Qt.WA_DeleteOnClose)

        print("Good bye")
        return super(Window, self).closeEvent(event)

    def keyPressEvent(self, event):
        modifiers = event.modifiers()
        ctrl_pressed = QtCore.Qt.ControlModifier & modifiers

        # Grouping subsets on pressing Ctrl + G
        if (ctrl_pressed and event.key() == QtCore.Qt.Key_G and
                not event.isAutoRepeat()):
            self.show_grouping_dialog()
            return

        super(Window, self).keyPressEvent(event)
        event.setAccepted(True)  # Avoid interfering other widgets

    def show_grouping_dialog(self):
        subsets = self.data["model"]["subsets"]
        if not subsets.is_groupable():
            self.echo("Grouping not enabled.")
            return

        selected = subsets.selected_subsets()
        if not selected:
            self.echo("No selected subset.")
            return

        dialog = SubsetGroupingDialog(items=selected, parent=self)
        dialog.grouped.connect(self._assetschanged)
        dialog.show()


class SubsetGroupingDialog(QtWidgets.QDialog):

    grouped = QtCore.Signal()

    def __init__(self, items, parent=None):
        super(SubsetGroupingDialog, self).__init__(parent=parent)
        self.setWindowTitle("Grouping Subsets")
        self.setMinimumWidth(250)
        self.setModal(True)

        self.items = items
        self.subsets = parent.data["model"]["subsets"]
        self.asset_id = parent.data["state"]["context"]["assetId"]

        name = QtWidgets.QLineEdit()
        name.setPlaceholderText("Remain blank to ungroup..")

        # Menu for pre-defined subset groups
        name_button = QtWidgets.QPushButton()
        name_button.setFixedWidth(18)
        name_button.setFixedHeight(20)
        name_menu = QtWidgets.QMenu(name_button)
        name_button.setMenu(name_menu)

        name_layout = QtWidgets.QHBoxLayout()
        name_layout.addWidget(name)
        name_layout.addWidget(name_button)
        name_layout.setContentsMargins(0, 0, 0, 0)

        group_btn = QtWidgets.QPushButton("Apply")

        layout = QtWidgets.QVBoxLayout(self)
        layout.addWidget(QtWidgets.QLabel("Group Name"))
        layout.addLayout(name_layout)
        layout.addWidget(group_btn)

        group_btn.clicked.connect(self.on_group)
        group_btn.setAutoDefault(True)
        group_btn.setDefault(True)

        self.name = name
        self.name_menu = name_menu

        self._build_menu()

    def _build_menu(self):
        menu = self.name_menu
        button = menu.parent()
        # Get and destroy the action group
        group = button.findChild(QtWidgets.QActionGroup)
        if group:
            group.deleteLater()

        active_groups = lib.get_active_group_config(self.asset_id,
                                                    include_predefined=True)
        # Build new action group
        group = QtWidgets.QActionGroup(button)
        for data in sorted(active_groups, key=lambda x: x["order"]):
            name = data["name"]
            icon = data["icon"]

            action = group.addAction(name)
            action.setIcon(icon)
            menu.addAction(action)

        group.triggered.connect(self._on_action_clicked)
        button.setEnabled(not menu.isEmpty())

    def _on_action_clicked(self, action):
        self.name.setText(action.text())

    def on_group(self):
        name = self.name.text().strip()
        self.subsets.group_subsets(name, self.asset_id, self.items)

        with lib.preserve_selection(tree_view=self.subsets.view,
                                    current_index=False):
            self.grouped.emit()
            self.close()


def show(debug=False, parent=None, use_context=False):
    """Display Loader GUI

    Arguments:
        debug (bool, optional): Run loader in debug-mode,
            defaults to False
        parent (QtCore.QObject, optional): The Qt object to parent to.
        use_context (bool): Whether to apply the current context upon launch

    """

    # Remember window
    if module.window is not None:
        try:
            module.window.show()

            # If the window is minimized then unminimize it.
            if module.window.windowState() & QtCore.Qt.WindowMinimized:
                module.window.setWindowState(QtCore.Qt.WindowActive)

            # Raise and activate the window
            module.window.raise_()             # for MacOS
            module.window.activateWindow()     # for Windows
            module.window.refresh()
            return
        except RuntimeError as e:
            if not str(e).rstrip().endswith("already deleted."):
                raise

            # Garbage collected
            module.window = None

    if debug:
        import traceback
        sys.excepthook = lambda typ, val, tb: traceback.print_last()

    with lib.application():

        # TODO: Global state, remove these
        lib.refresh_family_config_cache()
        lib.refresh_group_config_cache()

        window = Window(parent)
        window.show()
        window.setStyleSheet(style.load_stylesheet())

        if use_context:
            context = {"asset": api.Session["AVALON_ASSET"],
                       "silo": api.Session["AVALON_SILO"]}
            window.set_context(context, refresh=True)
        else:
            window.refresh()

        module.window = window

        # Pull window to the front.
        module.window.raise_()
        module.window.activateWindow()


def cli(args):

    import argparse

    parser = argparse.ArgumentParser()
    parser.add_argument("project")

    args = parser.parse_args(args)
    project = args.project

    io.install()

    # Store settings
    api.Session["AVALON_PROJECT"] = project

    from avalon import pipeline

    # Find the set config
    _config = pipeline.find_config()
    if hasattr(_config, "install"):
        _config.install()
    else:
        print("Config `%s` has no function `install`" %
              _config.__name__)

    show()



from .pipeline import (
    install,
    uninstall,

    Creator,

    ls,
    containerise,

)

from .workio import (
    open_file,
    save_file,
    current_file,
    has_unsaved_changes,
    file_extensions,
    work_root
)

from .lib import (
    lsattr,
    lsattrs,
    read,

    maintained_selection,
    unique_name
)


__all__ = [
    "install",
    "uninstall",

    "Creator",

    "ls",
    "containerise",

    # Workfiles API
    "open_file",
    "save_file",
    "current_file",
    "has_unsaved_changes",
    "file_extensions",
    "work_root",

    # Utility functions
    "lsattr",
    "lsattrs",
    "read",

    "maintained_selection",
    "unique_name"
]

# Backwards API compatibility
open = open_file
save = save_file



"""The safe production pipeline"""

import os
from setuptools import setup, find_packages
from avalon.version import version

# Git is required for deployment
assert len(version.split(".")) == 3, (
    "Could not compute patch version, make sure `git` is\n"
    "available and see version.py for details")

classifiers = [
    "Development Status :: 5 - Production/Stable",
    "License :: OSI Approved :: MIT License",
    "Intended Audience :: Developers",
    "Operating System :: OS Independent",
    "Programming Language :: Python",
    "Programming Language :: Python :: 2",
    "Programming Language :: Python :: 2.7",
    "Programming Language :: Python :: 3",
    "Programming Language :: Python :: 3.6",
    "Programming Language :: Python :: 3.7",
    "Topic :: Utilities",
    "Topic :: Software Development",
    "Topic :: Software Development :: Libraries :: Python Modules",
]

# Serialise version from Git revision to patch number
dirname = os.path.dirname(__file__)
fname = os.path.join(dirname, "avalon", "__version__.py")
with open(fname, "w") as f:
    f.write("version = \"%s\"\n" % version)

try:
    setup(
        name="avalon-core",
        version=version,
        description=__doc__,
        long_description=__doc__,
        url="https://github.com/getavalon/core",
        author="Marcus Ottosson",
        author_email="konstruktion@gmail.com",
        license="MIT",
        zip_safe=False,
        packages=find_packages(),
        package_data={
            "avalon": [
                "schema/*.json",
                "style/*.qrc",
                "style/*.qss",
                "style/rc/*.png",
                "fonts/opensans/*.txt",
                "fonts/opensans/*.ttf",
                "vendor/jsonschema/schemas/*.json",
                "vendor/qtawesome/fonts/*.ttf",
                "vendor/qtawesome/fonts/*.json",
            ]
        },
        classifiers=classifiers,
        install_requires=[
            "pymongo>=3.4,<4",
        ],
        python_requires=">2.7, <4",
    )

finally:
    # Clean-up
    os.remove(fname)



import os
import platform


HARDLINK = 0


def _create_windows(source, destination, link_type):
    """Creates hardlink at destination from source in Windows."""

    if link_type == HARDLINK:
        import ctypes
        from ctypes.wintypes import BOOL
        CreateHardLink = ctypes.windll.kernel32.CreateHardLinkW
        CreateHardLink.argtypes = [
            ctypes.c_wchar_p, ctypes.c_wchar_p, ctypes.c_void_p
        ]
        CreateHardLink.restype = BOOL

        res = CreateHardLink(destination, source, None)
        if res == 0:
            raise ctypes.WinError()
    else:
        raise NotImplementedError("Link type unrecognized.")


def _create_linux(source, destination, link_type):
    """Creates hardlink at destination from source in Linux."""

    if link_type == HARDLINK:
        os.link(source, destination)
    else:
        raise NotImplementedError("Link type unrecognized.")


def _create_osx(source, destination, link_type):
    """Creates hardlink at destination from source in OSX."""
    # OSX is not supported yet, as testing on Travis does not support Python
    # natively.
    raise NotImplementedError("OSX is not support yet.")


def create(source, destination, link_type=HARDLINK):
    """Creates a hardlink at destination referring to the same file."""

    system = platform.system()

    if system == "Windows":
        _create_windows(source, destination, link_type)
    elif system == "Linux":
        _create_linux(source, destination, link_type)
    elif system == "Darwin":
        _create_osx(source, destination, link_type)
    else:
        raise NotImplementedError(
            "Unsupported platform: \"{0}\"".format(system)
        )



# :coding: utf-8
# :copyright: Copyright (c) 2013 Martin Pengelly-Phillips
# :license: See LICENSE.txt.

'''Custom error classes.'''


class CollectionError(Exception):
    '''Raise when a collection error occurs.'''




"""Host API required Work Files tool"""
import sys
import os


def file_extensions():
    return [".comp"]


def has_unsaved_changes():
    from avalon.fusion.pipeline import get_current_comp

    comp = get_current_comp()
    return comp.GetAttrs()["COMPB_Modified"]


def save_file(filepath):
    from avalon.fusion.pipeline import get_current_comp

    comp = get_current_comp()
    comp.Save(filepath)


def open_file(filepath):
    # Hack to get fusion, see avalon.fusion.pipeline.get_current_comp()
    fusion = getattr(sys.modules["__main__"], "fusion", None)

    return fusion.LoadComp(filepath)


def current_file():
    from avalon.fusion.pipeline import get_current_comp

    comp = get_current_comp()
    current_filepath = comp.GetAttrs()["COMPS_FileName"]
    if not current_filepath:
        return None

    return current_filepath


def work_root(session):

    work_dir = session["AVALON_WORKDIR"]
    scene_dir = session.get("AVALON_SCENEDIR")
    if scene_dir:
        return os.path.join(work_dir, scene_dir)
    else:
        return work_dir



import sys

# Backwards compatibility
# Support `from avalon.tools import cbloader`
from . import (
    sceneinventory as cbsceneinventory,
    loader as cbloader,
)

from .lib import (
    application,
)

# Support `import avalon.tools.cbloader`
sys.modules[__name__ + ".cbsceneinventory"] = cbsceneinventory
sys.modules[__name__ + ".cbloader"] = cbloader

__all__ = [
    "application",

    "cbsceneinventory",
    "cbloader",
]



"""Initialize the avalon style."""
import logging
import platform
import os

from ..vendor.Qt import (
    QtCore,
    QtGui,
    __binding__
)

from . import colors

_log = logging.getLogger("avalon.style")


def load_stylesheet():
    """Loads the stylesheet.

    This takes care of importing the rc module for the correct binding loaded
    with Qt.py.

    Returns:
        str: The Stylesheet string.

    """

    # Smart import of the rc file
    if __binding__ == "PySide":
        from . import pyside_style_rc
    elif __binding__ == "PyQt4":
        from . import pyqt_style_rc
    elif __binding__ == "PySide2":
        from . import pyside_style_rc
    elif __binding__ == "PyQt5":
        from . import pyqt5_style_rc
    else:
        RuntimeError("Unsupported Qt binding: {0}".format(__binding__))

    # Load the fonts
    _load_fonts()

    # Load the stylesheet content from resources
    path = os.path.join(os.path.dirname(__file__), "style.qss")
    if not os.path.exists(path):
        _log.error("Unable to load stylesheet, file not found in resources")
        return ""

    with open(path, "r") as f:
        stylesheet = f.read()

        # fix bug on mac os; see cbdarkstyle issue #12 on github
        if platform.system().lower() == 'darwin':
            mac_fix = '''
            QDockWidget::title
            {
                background-color: #353434;
                text-align: center;
                height: 12px;
            }
            '''
            stylesheet += mac_fix

    return stylesheet


def _load_fonts():
    """Load the fonts.

    This will add the OpenSans fonts to the QFontDatabase.

    Returns:
        list: The registered font ids.

    """

    font_root = os.path.join(os.path.dirname(__file__), "fonts")
    fonts = [
        "opensans/OpenSans-Bold.ttf",
        "opensans/OpenSans-BoldItalic.ttf",
        "opensans/OpenSans-ExtraBold.ttf",
        "opensans/OpenSans-ExtraBoldItalic.ttf",
        "opensans/OpenSans-Italic.ttf",
        "opensans/OpenSans-Light.ttf",
        "opensans/OpenSans-LightItalic.ttf",
        "opensans/OpenSans-Regular.ttf",
        "opensans/OpenSans-Semibold.ttf",
        "opensans/OpenSans-SemiboldItalic.ttf",
    ]

    for font in fonts:
        path = os.path.join(font_root, font)
        QtGui.QFontDatabase.addApplicationFont(path)



import datetime
import pprint
import inspect

from ...vendor.Qt import QtWidgets, QtCore, QtCompat
from ...vendor import qtawesome
from ... import io
from ... import api
from ... import pipeline

from .. import lib as tools_lib
from ..delegates import VersionDelegate
from ..widgets import OptionalMenu, OptionalAction, OptionDialog

from .model import (
    SubsetsModel,
    SubsetFilterProxyModel,
    FamiliesFilterProxyModel,
)
from ..delegates import PrettyTimeDelegate


class SubsetWidget(QtWidgets.QWidget):
    """A widget that lists the published subsets for an asset"""

    active_changed = QtCore.Signal()    # active index changed
    version_changed = QtCore.Signal()   # version state changed for a subset

    def __init__(self, enable_grouping=True, parent=None):
        super(SubsetWidget, self).__init__(parent=parent)

        model = SubsetsModel(grouping=enable_grouping)
        proxy = SubsetFilterProxyModel()
        family_proxy = FamiliesFilterProxyModel()
        family_proxy.setSourceModel(proxy)

        filter = QtWidgets.QLineEdit()
        filter.setPlaceholderText("Filter subsets..")

        groupable = QtWidgets.QCheckBox("Enable Grouping")
        groupable.setChecked(enable_grouping)

        top_bar_layout = QtWidgets.QHBoxLayout()
        top_bar_layout.addWidget(filter)
        top_bar_layout.addWidget(groupable)

        view = QtWidgets.QTreeView()
        view.setIndentation(20)
        view.setStyleSheet("""
            QTreeView::item{
                padding: 5px 1px;
                border: 0px;
            }
        """)
        view.setAllColumnsShowFocus(True)

        # Set view delegates
        version_delegate = VersionDelegate()
        column = model.Columns.index("version")
        view.setItemDelegateForColumn(column, version_delegate)

        time_delegate = PrettyTimeDelegate()
        column = model.Columns.index("time")
        view.setItemDelegateForColumn(column, time_delegate)

        layout = QtWidgets.QVBoxLayout(self)
        layout.setContentsMargins(0, 0, 0, 0)
        layout.addLayout(top_bar_layout)
        layout.addWidget(view)

        view.setContextMenuPolicy(QtCore.Qt.CustomContextMenu)
        view.setSelectionMode(QtWidgets.QAbstractItemView.ExtendedSelection)
        view.setSortingEnabled(True)
        view.sortByColumn(1, QtCore.Qt.AscendingOrder)
        view.setAlternatingRowColors(True)

        self.data = {
            "delegates": {
                "version": version_delegate,
                "time": time_delegate
            },
            "state": {
                "groupable": groupable
            }
        }

        self.proxy = proxy
        self.model = model
        self.view = view
        self.filter = filter
        self.family_proxy = family_proxy

        # settings and connections
        self.proxy.setSourceModel(self.model)
        self.proxy.setDynamicSortFilter(True)
        self.proxy.setFilterCaseSensitivity(QtCore.Qt.CaseInsensitive)

        self.view.setModel(self.family_proxy)
        self.view.customContextMenuRequested.connect(self.on_context_menu)

        header = self.view.header()
        # Enforce the columns to fit the data (purely cosmetic)
        QtCompat.setSectionResizeMode(
            header, QtWidgets.QHeaderView.ResizeToContents)

        selection = view.selectionModel()
        selection.selectionChanged.connect(self.active_changed)

        version_delegate.version_changed.connect(self.version_changed)

        groupable.stateChanged.connect(self.set_grouping)

        self.filter.textChanged.connect(self.proxy.setFilterRegExp)
        self.filter.textChanged.connect(self.view.expandAll)

        self.model.refresh()

        # Expose this from the widget as a method
        self.set_family_filters = self.family_proxy.setFamiliesFilter

    def is_groupable(self):
        return self.data["state"]["groupable"].checkState()

    def set_grouping(self, state):
        with tools_lib.preserve_selection(tree_view=self.view,
                                          current_index=False):
            self.model.set_grouping(state)

    def on_context_menu(self, point):

        point_index = self.view.indexAt(point)
        if not point_index.isValid():
            return

        node = point_index.data(self.model.ItemRole)
        if node.get("isGroup"):
            return

        # Get all representation->loader combinations available for the
        # index under the cursor, so we can list the user the options.
        available_loaders = api.discover(api.Loader)
        loaders = list()

        version_id = node["version_document"]["_id"]
        representations = io.find({"type": "representation",
                                   "parent": version_id})
        for representation in representations:
            for loader in api.loaders_from_representation(
                    available_loaders,
                    representation["_id"]
            ):
                loaders.append((representation, loader))

        if not loaders:
            # no loaders available
            self.echo("No compatible loaders available for this version.")
            return

        # Get selected rows
        selection = self.view.selectionModel()
        rows = selection.selectedRows(column=0)
        # Ensure active point index is also used as first column so we can
        # correctly push it to the end in the rows list.
        point_index = point_index.sibling(point_index.row(), 0)
        # Ensure point index is run first.
        try:
            rows.remove(point_index)
        except ValueError:
            pass
        rows.insert(0, point_index)

        # Enable optional action when only one item being selected
        enable_option = len(rows) == 1

        def sorter(value):
            """Sort the Loaders by their order and then their name"""
            Plugin = value[1]
            return Plugin.order, Plugin.__name__

        # List the available loaders
        menu = OptionalMenu(self)
        for representation, loader in sorted(loaders, key=sorter):

            # Label
            label = getattr(loader, "label", None)
            if label is None:
                label = loader.__name__

            # Add the representation as suffix
            label = "{0} ({1})".format(label, representation["name"])

            # Support font-awesome icons using the `.icon` and `.color`
            # attributes on plug-ins.
            icon = getattr(loader, "icon", None)
            if icon is not None:
                try:
                    key = "fa.{0}".format(icon)
                    color = getattr(loader, "color", "white")
                    icon = qtawesome.icon(key, color=color)
                except Exception as e:
                    print("Unable to set icon for loader "
                          "{}: {}".format(loader, e))
                    icon = None

            # Optional action
            use_option = enable_option and hasattr(loader, "options")
            action = OptionalAction(label, icon, use_option, menu)

            if use_option:
                # Add option box tip
                action.set_option_tip(loader.options)

            action.setData((representation, loader))

            # Add tooltip and statustip from Loader docstring
            tip = inspect.getdoc(loader)
            if tip:
                action.setToolTip(tip)
                action.setStatusTip(tip)

            menu.addAction(action)

        # Show the context action menu
        global_point = self.view.mapToGlobal(point)
        action = menu.exec_(global_point)
        if not action:
            return

        # Find the representation name and loader to trigger
        action_representation, loader = action.data()
        representation_name = action_representation["name"]  # extension
        options = None

        # Pop option dialog
        if getattr(action, "optioned", False):
            dialog = OptionDialog(self)
            dialog.setWindowTitle(action.label + " Options")
            dialog.create(loader.options)

            if not dialog.exec_():
                return

            # Get option
            options = dialog.parse()

        # Run the loader for all selected indices, for those that have the
        # same representation available

        # Trigger
        for row in rows:
            node = row.data(self.model.ItemRole)
            if node.get("isGroup"):
                continue

            version_id = node["version_document"]["_id"]
            representation = io.find_one({"type": "representation",
                                          "name": representation_name,
                                          "parent": version_id})
            if not representation:
                self.echo("Subset '{}' has no representation '{}'".format(
                          node["subset"],
                          representation_name
                          ))
                continue

            try:
                api.load(Loader=loader,
                         representation=representation,
                         options=options)

            except pipeline.IncompatibleLoaderError as exc:
                self.echo(exc)
                continue

    def selected_subsets(self):
        selection = self.view.selectionModel()
        rows = selection.selectedRows(column=0)

        subsets = list()
        for row in rows:
            node = row.data(self.model.ItemRole)
            if not node.get("isGroup"):
                subsets.append(node)

        return subsets

    def group_subsets(self, name, asset_id, nodes):
        field = "data.subsetGroup"

        if name:
            update = {"$set": {field: name}}
            self.echo("Group subsets to '%s'.." % name)
        else:
            update = {"$unset": {field: ""}}
            self.echo("Ungroup subsets..")

        subsets = list()
        for node in nodes:
            subsets.append(node["subset"])

        filter = {
            "type": "subset",
            "parent": asset_id,
            "name": {"$in": subsets},
        }
        io.update_many(filter, update)

    def echo(self, message):
        print(message)


class VersionTextEdit(QtWidgets.QTextEdit):
    """QTextEdit that displays version specific information.

    This also overrides the context menu to add actions like copying
    source path to clipboard or copying the raw data of the version
    to clipboard.

    """
    def __init__(self, parent=None):
        super(VersionTextEdit, self).__init__(parent=parent)

        self.data = {
            "source": None,
            "raw": None
        }

        # Reset
        self.set_version(None)

    def set_version(self, version_id):

        if not version_id:
            # Reset state to empty
            self.data = {
                "source": None,
                "raw": None,
            }
            self.setText("")
            self.setEnabled(True)
            return

        self.setEnabled(True)

        print("Querying..")

        version = io.find_one({"_id": version_id, "type": "version"})
        assert version, "Not a valid version id"

        subset = io.find_one({"_id": version["parent"], "type": "subset"})
        assert subset, "No valid subset parent for version"

        # Define readable creation timestamp
        created = version["data"]["time"]
        created = datetime.datetime.strptime(created, "%Y%m%dT%H%M%SZ")
        created = datetime.datetime.strftime(created, "%b %d %Y %H:%M")

        comment = version["data"].get("comment", None) or "No comment"

        source = version["data"].get("source", None)
        source_label = source if source else "No source"

        # Store source and raw data
        self.data["source"] = source
        self.data["raw"] = version

        data = {
            "subset": subset["name"],
            "version": version["name"],
            "comment": comment,
            "created": created,
            "source": source_label
        }

        self.setHtml(u"""
<h3>{subset} v{version:03d}</h3>
<b>Comment</b><br>
{comment}<br>
<br>
<b>Created</b><br>
{created}<br>
<br>
<b>Source</b><br>
{source}<br>""".format(**data))

    def contextMenuEvent(self, event):
        """Context menu with additional actions"""
        menu = self.createStandardContextMenu()

        # Add additional actions when any text so we can assume
        # the version is set.
        if self.toPlainText().strip():

            menu.addSeparator()
            action = QtWidgets.QAction("Copy source path to clipboard",
                                       menu)
            action.triggered.connect(self.on_copy_source)
            menu.addAction(action)

            action = QtWidgets.QAction("Copy raw data to clipboard",
                                       menu)
            action.triggered.connect(self.on_copy_raw)
            menu.addAction(action)

        menu.exec_(event.globalPos())
        del menu

    def on_copy_source(self):
        """Copy formatted source path to clipboard"""
        source = self.data.get("source", None)
        if not source:
            return

        path = source.format(root=api.registered_root())
        clipboard = QtWidgets.QApplication.clipboard()
        clipboard.setText(path)

    def on_copy_raw(self):
        """Copy raw version data to clipboard

        The data is string formatted with `pprint.pformat`.

        """
        raw = self.data.get("raw", None)
        if not raw:
            return

        raw_text = pprint.pformat(raw)
        clipboard = QtWidgets.QApplication.clipboard()
        clipboard.setText(raw_text)


class VersionWidget(QtWidgets.QWidget):
    """A Widget that display information about a specific version"""
    def __init__(self, parent=None):
        super(VersionWidget, self).__init__(parent=parent)

        layout = QtWidgets.QVBoxLayout(self)

        label = QtWidgets.QLabel("Version")
        data = VersionTextEdit()
        data.setReadOnly(True)
        layout.addWidget(label)
        layout.addWidget(data)

        self.data = data

    def set_version(self, version_id):
        self.data.set_version(version_id)


class FamilyListWidget(QtWidgets.QListWidget):
    """A Widget that lists all available families"""

    NameRole = QtCore.Qt.UserRole + 1
    active_changed = QtCore.Signal(list)

    def __init__(self, parent=None):
        super(FamilyListWidget, self).__init__(parent=parent)

        multi_select = QtWidgets.QAbstractItemView.ExtendedSelection
        self.setSelectionMode(multi_select)
        self.setAlternatingRowColors(True)
        # Enable RMB menu
        self.setContextMenuPolicy(QtCore.Qt.CustomContextMenu)
        self.customContextMenuRequested.connect(self.show_right_mouse_menu)

        self.itemChanged.connect(self._on_item_changed)

    def refresh(self):
        """Refresh the listed families.

        This gets all unique families and adds them as checkable items to
        the list.

        """

        family = io.distinct("data.family")
        families = io.distinct("data.families")
        unique_families = list(set(family + families))

        # Rebuild list
        self.blockSignals(True)
        self.clear()
        for name in sorted(unique_families):

            family = tools_lib.get_family_cached_config(name)
            if family.get("hideFilter"):
                continue

            label = family.get("label", name)
            icon = family.get("icon", None)

            # TODO: This should be more managable by the artist
            # Temporarily implement support for a default state in the project
            # configuration
            state = family.get("state", True)
            state = QtCore.Qt.Checked if state else QtCore.Qt.Unchecked

            item = QtWidgets.QListWidgetItem(parent=self)
            item.setText(label)
            item.setFlags(item.flags() | QtCore.Qt.ItemIsUserCheckable)
            item.setData(self.NameRole, name)
            item.setCheckState(state)

            if icon:
                item.setIcon(icon)

            self.addItem(item)
        self.blockSignals(False)

        self.active_changed.emit(self.get_filters())

    def get_filters(self):
        """Return the checked family items"""

        items = [self.item(i) for i in
                 range(self.count())]

        return [item.data(self.NameRole) for item in items if
                item.checkState() == QtCore.Qt.Checked]

    def _on_item_changed(self):
        self.active_changed.emit(self.get_filters())

    def _set_checkstate_all(self, state):
        _state = QtCore.Qt.Checked if state is True else QtCore.Qt.Unchecked
        self.blockSignals(True)
        for i in range(self.count()):
            item = self.item(i)
            item.setCheckState(_state)
        self.blockSignals(False)
        self.active_changed.emit(self.get_filters())

    def show_right_mouse_menu(self, pos):
        """Build RMB menu under mouse at current position (within widget)"""

        # Get mouse position
        globalpos = self.viewport().mapToGlobal(pos)

        menu = QtWidgets.QMenu(self)

        # Add enable all action
        state_checked = QtWidgets.QAction(menu, text="Enable All")
        state_checked.triggered.connect(
            lambda: self._set_checkstate_all(True))
        # Add disable all action
        state_unchecked = QtWidgets.QAction(menu, text="Disable All")
        state_unchecked.triggered.connect(
            lambda: self._set_checkstate_all(False))

        menu.addAction(state_checked)
        menu.addAction(state_unchecked)

        menu.exec_(globalpos)
